///////////////////////////////////////////////////////////////////////////////
// File: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2009 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "SocketServer.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/IQueueTimers.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/IAllocateStreamSockets.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::Output;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::CException;
using JetByteTools::Core::GetStringLength;
using JetByteTools::Core::IQueueTimers;
using JetByteTools::Core::IIndexedOpaqueUserData;

using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IIOPool;

using JetByteTools::Socket::CStreamSocketServerEx;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ILimitConnections;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::NoZeroByteRead;
using JetByteTools::Socket::CSmartStreamSocket;

using std::string;

///////////////////////////////////////////////////////////////////////////////
// CSocketServer
///////////////////////////////////////////////////////////////////////////////

CSocketServer::CSocketServer(
   const string &welcomeMessage,
   IQueueTimers &timerQueue,
   const IFullAddress &address,
   const ListenBacklog listenBacklog,
   IIOPool &pool,
   IAllocateStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   ILimitConnections &connectionLimiter)
   :  CStreamSocketServerEx(address, listenBacklog, *this, pool, socketAllocator, bufferAllocator, NoZeroByteRead, connectionLimiter),
      m_welcomeMessage(welcomeMessage),
      m_timerQueue(timerQueue),
      m_userDataIndex(socketAllocator.RequestUserDataSlot(_T("CSocketServer::UserData")))
{

}

CSocketServer::~CSocketServer()
{
   WaitForShutdownToComplete();
}

void CSocketServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   Output(_T("OnConnectionEstablished"));

   IQueueTimers::Handle timerHandle = m_timerQueue.CreateTimer();

   socket.SetUserData(m_userDataIndex, timerHandle);

   if (socket.TryWrite(m_welcomeMessage.c_str(), GetStringLength<DWORD>(m_welcomeMessage)))
   {
      socket.TryRead();
   }

   SetTimer(socket);
}

void CSocketServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   EchoMessage(socket, buffer);

   socket.TryRead();
}

void CSocketServer::OnConnectionClientClose(
    IStreamSocket &socket)
{
   Output(_T("OnConnectionClientClose"));

   IQueueTimers::Handle timerHandle = socket.GetUserData(m_userDataIndex);

   m_timerQueue.CancelTimerWithRefCountedUserData(timerHandle, &socket);
}

void CSocketServer::OnConnectionReset(
    IStreamSocket &socket,
    const DWORD /*lastError*/)
{
   Output(_T("OnConnectionReset"));

   IQueueTimers::Handle timerHandle = socket.GetUserData(m_userDataIndex);

   m_timerQueue.CancelTimerWithRefCountedUserData(timerHandle, &socket);
}

void CSocketServer::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   IQueueTimers::Handle timerHandle = userData.GetUserData(m_userDataIndex);

   m_timerQueue.DestroyTimer(timerHandle);
}

void CSocketServer::OnTimer(
   UserData userData)
{
   Output(_T("OnTimer"));

   CSmartStreamSocket socket(*reinterpret_cast<IStreamSocket *>(userData), CSmartStreamSocket::TakeOwnershipOfExistingReference);

   const string message("Periodic per connection message\r\n");

   socket->Write(message.c_str(), GetStringLength<DWORD>(message));

   SetTimer(socket.GetRef());
}

void CSocketServer::SetTimer(
   IStreamSocket &socket)
{
   IQueueTimers::Handle timerHandle = socket.GetUserData(m_userDataIndex);

   static const Milliseconds timeout = 2000;  // 2 seconds

   // Timer queue will manage reference required on socket.

   m_timerQueue.SetTimerWithRefCountedUserData(timerHandle, *this, timeout, &socket);
}

void CSocketServer::EchoMessage(
   IStreamSocket &socket,
   CSmartBuffer &buffer) const
{
   //DEBUG_ONLY(Output(_T("Data Echoed -\r\n") + DumpData(buffer->GetMemory(), buffer->GetUsed(), 60, true)));

   socket.TryWrite(buffer);
}

///////////////////////////////////////////////////////////////////////////////
// End of file: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
