///////////////////////////////////////////////////////////////////////////////
// File: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2009 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "SocketServer.h"
#include "ThreadPool.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/Win32Tools/Win32Exception.h"

#include "JetByteTools/IOTools/SmartBuffer.h"
#include "JetByteTools/IOTools/Buffer.h"

#include "JetByteTools/SocketTools/StreamSocketBroadcastableConnectionCollection.h"
#include "JetByteTools/SocketTools/IStreamSocket.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::ToString;
using JetByteTools::Core::CException;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::GetStringLength;
using JetByteTools::Core::IQueueTimers;

using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IBuffer;
using JetByteTools::IO::IIOPool;

using JetByteTools::Socket::CStreamSocketServerEx;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ILimitConnections;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::CStreamSocketBroadcastableConnectionCollection;
using JetByteTools::Socket::NoZeroByteRead;
using JetByteTools::Socket::CSmartStreamSocket;

using std::string;

///////////////////////////////////////////////////////////////////////////////
// CSocketServer
///////////////////////////////////////////////////////////////////////////////

CSocketServer::CSocketServer(
   IQueueTimers &timerQueue,
   CStreamSocketBroadcastableConnectionCollection &connections,
   CThreadPool &threadPool,
   const IFullAddress &address,
   const ListenBacklog listenBacklog,
   IIOPool &pool,
   IAllocateStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   ILimitConnections &connectionLimiter)
   :  CStreamSocketServerEx(address, listenBacklog, *this, pool, socketAllocator, bufferAllocator, NoZeroByteRead, connectionLimiter),
      m_timerQueue(timerQueue),
      m_timerHandle1(timerQueue.CreateTimer()),
      m_timerHandle2(timerQueue.CreateTimer()),
      m_connections(connections),
      m_pool(threadPool),
      m_timersCanBeSet(true)
{
   SetTimer1();
   SetTimer2();
}

CSocketServer::~CSocketServer()
{
   m_timerQueue.DestroyTimer(m_timerHandle1);
   m_timerQueue.DestroyTimer(m_timerHandle2);

   WaitForShutdownToComplete();
}

void CSocketServer::SetTimer1()
{
   if (m_timersCanBeSet)
   {
      // Set a timer for 3 seconds...

      m_timerQueue.SetTimer(m_timerHandle1, *this, 3000, 1);
   }
}

void CSocketServer::SetTimer2()
{
   if (m_timersCanBeSet)
   {
      // Set a timer for 6 seconds...

      m_timerQueue.SetTimer(m_timerHandle2, *this, 6000, 2);
   }
}

void CSocketServer::StopTimers()
{
   m_timersCanBeSet = false;

   m_timerQueue.CancelTimer(m_timerHandle1);
   m_timerQueue.CancelTimer(m_timerHandle2);
}

void CSocketServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress &address)
{
   Output(_T("OnConnectionEstablished"));

   m_connections.AddConnection(socket);

   m_pool.DispatchConnectionEstablished(socket, address);
}

void CSocketServer::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   m_pool.OnSocketReleased(userData);
}

void CSocketServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   ProcessDataStream(socket, buffer);

   socket.TryRead(buffer);
}

void CSocketServer::OnConnectionClientClose(
    IStreamSocket &socket)
{
   m_connections.RemoveConnection(socket);

   m_pool.OnConnectionClientClose(socket);
}

void CSocketServer::OnConnectionReset(
    IStreamSocket &socket,
    const DWORD lastError)
{
   m_connections.RemoveConnection(socket);

   m_pool.OnConnectionReset(socket, lastError);
}

void CSocketServer::OnTimer(
   UserData userData)
{
   // We the regular work to the business logic thread pool
   // to have it done there, so that we don't do too much
   // work on the timer queue thread.

   m_pool.DispatchServerTimer(static_cast<DWORD>(userData));

   // It may be more desirable to set these timers again after the work
   // item that you've just dispatched has been completed by the thread pool.

   if (userData == 1)
   {
      SetTimer1();
   }
   else
   {
      SetTimer2();
   }
}

void CSocketServer::ProcessDataStream(
   IStreamSocket &socket,
   CSmartBuffer &buffer) const
{
   bool done;

   //DEBUG_ONLY(Output(_T("ProcessDataStream:\r\n") + DumpData(buffer->GetMemory(), buffer->GetUsed(), 60, true)));

   do
   {
      done = true;

      const IBuffer::BufferSize used = buffer->GetUsed();

      if (used >= GetMinimumMessageSize())
      {
         const IBuffer::BufferSize messageSize = GetMessageSize(buffer.GetRef());

         if (messageSize == 0)
         {
            // havent got a complete message yet.

            // we null terminate our messages in the buffer, so we need to reserve
            // a byte of the buffer for this purpose...

            if (used == (buffer->GetSize() - 1))
            {
               Output(_T("Too much data!"));

               const string response("-ERR too much data! Go away!\r\n");

               // Write this message and then shutdown the sending side of the socket.
               socket.Write(response.c_str(), GetStringLength<DWORD>(response));
               socket.Shutdown();

               // throw the rubbish away
               buffer->Empty();

               done = true;
            }
         }
         else if (used == messageSize)
         {
            Output(_T("Got complete, distinct, message"));
            // we have a whole, distinct, message

            buffer->AddData(0);   // null terminate the command string;

            ProcessCommand(socket, buffer);

            done = true;
         }
         else if (used > messageSize)
         {
            Output(_T("Got message plus extra data"));
            // we have a message, plus some more data

            // allocate a new buffer, copy the data we want into it and then
            // read some more into the original buffer...

            CSmartBuffer message(buffer->SplitBufferFromFront(messageSize));

            message->AddData(0);   // null terminate the command string;

            ProcessCommand(socket, message);

            // loop again, we may have another complete message in there...

            done = false;
         }
      }
   }
   while (!done);
}

IBuffer::BufferSize CSocketServer::GetMinimumMessageSize() const
{
   // The smallest possible command we accept is TOP (plus the crlf terminator,
   // once we have this many bytes we can start with try and work out
   // what we have...
   return 5;
}

IBuffer::BufferSize CSocketServer::GetMessageSize(
   const IBuffer &buffer) const
{
   const BYTE *pData = buffer.GetMemory();

   const IBuffer::BufferSize used = buffer.GetUsed();

   for (IBuffer::BufferSize i = 0; i < used; ++i)
   {
      if (pData[i] == '\r')
      {
         if (i + 1 < used && pData[i + 1] == '\n')
         {
            // the end of the message is i+1
            // we actually want a count of characters, not a zero based
            // index, so we have to add 1...
            return i + 1 + 1;
         }
      }
   }

   return 0;
}

void CSocketServer::ProcessCommand(
   IStreamSocket &socket,
   CSmartBuffer &buffer) const
{
   m_pool.DispatchReadCompleted(socket, buffer);
}

///////////////////////////////////////////////////////////////////////////////
// End of file: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
