///////////////////////////////////////////////////////////////////////////////
// File: ThreadPool.cpp
/////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2009 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "ThreadPool.h"

#include "JetByteTools/IOTools/IBuffer.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/IProvideUserData.h"
#include "JetByteTools/CoreTools/IIndexedOpaqueUserData.h"
#include "JetByteTools/CoreTools/StringConverter.h"

#include "JetByteTools/IOTools/SmartBuffer.h"
#include "JetByteTools/IOTools/IAllocateBuffers.h"

#include "JetByteTools/SocketTools/AddressRenderer.h"
#include "JetByteTools/SocketTools/SmartStreamSocket.h"
#include "JetByteTools/SocketTools/StreamSocketBroadcastableConnectionCollection.h"
#include "JetByteTools/SocketTools/IStreamSocket.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::CManualResetEvent;
using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CException;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::ToString;
using JetByteTools::Core::ToStringA;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::IProvideUserData;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::GetStringLength;
using JetByteTools::Core::IAsyncWorkItem;
using JetByteTools::Core::IQueueTimers;
using JetByteTools::Core::IThreadPoolWorkerThreadCallback;

using JetByteTools::IO::IBuffer;
using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::CSmartBuffer;

using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::CAddressRenderer;
using JetByteTools::Socket::CSmartStreamSocket;
using JetByteTools::Socket::CStreamSocketBroadcastableConnectionCollection;

using std::string;

///////////////////////////////////////////////////////////////////////////////
// CThreadPool
///////////////////////////////////////////////////////////////////////////////

CThreadPool::CThreadPool(
   IQueueTimers &timerQueue,
   CStreamSocketBroadcastableConnectionCollection &connections,
   IAllocateBuffers &bufferAllocator,
   IProvideUserData &userDataProvider,
   const ThreadCount initialThreads,
   const ThreadCount minThreads,
   const ThreadCount maxThreads,
   const ThreadCount maxDormantThreads,
   const Milliseconds poolMaintPeriod,
   const Milliseconds dispatchTimeout)
   :  m_threadPool(
         *this,
         initialThreads,
         minThreads,
         maxThreads,
         maxDormantThreads,
         poolMaintPeriod,
         dispatchTimeout,
         *this),
      m_timerQueue(timerQueue),
      m_connections(connections),
      m_bufferAllocator(bufferAllocator),
      m_userDataIndex(userDataProvider.RequestUserDataSlot(_T("CThreadPool::PerConnectionDataIndex")))
{
}

void CThreadPool::Start()
{
   m_threadPool.Start();
}

void CThreadPool::WaitForShutdownToComplete(
   const Milliseconds forceTerminationAfter)
{
   (void)forceTerminationAfter;
   // unix can't force thread termination at present ...

   m_threadPool.WaitForShutdownToComplete(/*forceTerminationAfter*/);
}

void CThreadPool::DispatchConnectionEstablished(
   IStreamSocket &socket,
   const IAddress &address)
{
   // Allocate per connection data

   CPerConnectionData *pData = new CPerConnectionData(CAddressRenderer::AsString(address, true), socket, m_timerQueue, *this);

   // And store our data in the socket...

   socket.SetUserPointer(m_userDataIndex, pData);

   DoDispatch(socket, 0, ConnectionEstablished);
}

void CThreadPool::DispatchReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   DoDispatch(socket, buffer.Detach(), ReadCompleted);
}

void CThreadPool::DispatchTimer(
   IStreamSocket &socket)
{
   DoDispatch(socket, 0, Timer);
}

void CThreadPool::DispatchServerTimer(
   const DWORD timerId)
{
   m_threadPool.Dispatch(
      static_cast<ULONG_PTR>(timerId),
      ServerTimer,
      0);
}

void CThreadPool::DoDispatch(
   IStreamSocket &socket,
   IBuffer *pBuffer,
   DispatchEvents event)
{
   socket.AddRef();

   if (!pBuffer)
   {
      if (event != ConnectionEstablished &&
          event != Timer)
      {
         throw CException(_T("CThreadPool::DoDispatch()"), _T("Unexpected: Only ConnectionEstablished and Timer events can have a null pBuffer"));
      }
   }

   m_threadPool.Dispatch(
      reinterpret_cast<ULONG_PTR>(&socket),
      event,
      pBuffer);
}

void CThreadPool::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   // release per connection data

   CPerConnectionData &data = GetPerConnectionData(userData);

   userData.SetUserPointer(m_userDataIndex, 0);

   delete &data;
}

CThreadPool::CPerConnectionData &CThreadPool::GetPerConnectionData(
   const IIndexedOpaqueUserData &userData) const
{
   return *static_cast<CPerConnectionData*>(userData.GetUserPointer(m_userDataIndex));
}

void CThreadPool::OnConnectionEstablished(
   IStreamSocket &socket)
{
   CPerConnectionData &data = GetPerConnectionData(socket);

   std::string welcomeMessage("+OK POP3 server ready - ");

   welcomeMessage = welcomeMessage + CStringConverter::TtoA(data.GetConnectionDetails()) + "\r\n";

   socket.Write(welcomeMessage.c_str(), GetStringLength<DWORD>(welcomeMessage));

   socket.TryRead();
}

void CThreadPool::ProcessMessage(
   IStreamSocket &socket,
   const CSmartBuffer &buffer) const
{
   std::string command = reinterpret_cast<const char*>(buffer->GetMemory());

   for (IBuffer::BufferSize  i = 0; i < command.length(); ++i)
   {
      command[i] = (char)toupper(command[i]);
   }

   bool ok = false;

   if (command.length() >= 5)
   {
      // cant be a valid command, unless it's more than 5 chars

      if (command.substr(0, 3) == "TOP")
      {
         ok = true;
      }
      else
      {
         const std::string cmd = command.substr(0, 4);

         if (cmd == "APOP" ||
             cmd == "DELE" ||
             cmd == "LIST" ||
             cmd == "NOOP" ||
             cmd == "PASS" ||
             cmd == "QUIT" ||
             cmd == "RETR" ||
             cmd == "RSET" ||
             cmd == "STAT" ||
             cmd == "UIDL" ||
             cmd == "USER")
         {
            ok = true;
         }
      }
   }

   CPerConnectionData &data = GetPerConnectionData(socket);

   data.UseConnection();

   if (ok)
   {
      // We understand, but we aren't really a POP3 server...

      std::string response("-ERR sorry, we understand what you want, but, we're just a fake POP3 server... ");

      response = response + CStringConverter::TtoA(data.GetConnectionDetails()) + "\r\n";

      // We take a while to reply because we do lots of processing...

      //::Sleep(5000);

      socket.Write(response.c_str(), GetStringLength<DWORD>(response));
   }
   else
   {
      std::string response("-ERR sorry, we dont know what you mean, bye... ");

      response = response + CStringConverter::TtoA(data.GetConnectionDetails()) + "\r\n";

      socket.Write(response.c_str(), GetStringLength<DWORD>(response));
      socket.Shutdown();
   }
}

void CThreadPool::OnConnectionClientClose(
    IStreamSocket &socket)
{
   CPerConnectionData &data = GetPerConnectionData(socket);

   data.CancelTimer();
}

void CThreadPool::OnConnectionReset(
    IStreamSocket &socket,
    const DWORD /*lastError*/)
{
   CPerConnectionData &data = GetPerConnectionData(socket);

   data.CancelTimer();
}

void CThreadPool::OnTimer(
   IStreamSocket &socket)
{
   CPerConnectionData &data = GetPerConnectionData(socket);

   const string periodicMessage("Every so often we send out periodic messages... Your details are: " + CStringConverter::TtoA(data.GetConnectionDetails()) + "\r\n");

   socket.Write(periodicMessage.c_str(), GetStringLength<DWORD>(periodicMessage));

   data.SetPeriodicTimer();
}

void CThreadPool::OnServerTimer(
   const DWORD timerId)
{
   CSmartBuffer buffer(m_bufferAllocator.Allocate());

   const string message = "Periodic server broadcast operation: " + ToStringA(timerId) + "\r\n";

   buffer->AddData(message.c_str(), GetStringLength<DWORD>(message));

   // Derive from one of the connection collection classes if you need to do clever stuff with
   // each connection, such as access their user data, or send connection specific information.
   // The broadcast loop should be the basis of how you transmit data to the collection...

   m_connections.Broadcast(buffer.GetRef());
}

IThreadPoolWorkerThreadCallback &CThreadPool::CreateWorkerThread()
{
   return *this;
}

void CThreadPool::DestroyWorkerThread(
   IThreadPoolWorkerThreadCallback & /*thread*/)
{

}

void CThreadPool::DisposeOfQueuedItemsAfterShutdown(
   const ULONG_PTR completionKey,
   const DWORD operation,
   IAsyncWorkItem *pWorkItem)
{
   Output(_T("DisposeOfQueuedItemsAfterShutdown"));

   if (operation != ServerTimer)
   {
      reinterpret_cast<IStreamSocket *>(completionKey)->Release();
   }

   if (operation == ReadCompleted)
   {
      static_cast<IBuffer *>(pWorkItem)->Release();
   }
}

bool CThreadPool::Initialise()
{
   Output(_T("Initialise"));

   return true;
}

void CThreadPool::Process(
   const ULONG_PTR completionKey,
   const DWORD operation,
   IAsyncWorkItem *pWorkItem)
{
   try
   {
      if (operation == ServerTimer)
      {
         OnServerTimer(static_cast<DWORD>(completionKey));
      }
      else
      {
         CSmartStreamSocket socket(reinterpret_cast<IStreamSocket *>(completionKey));

         try
         {
            switch(operation)
            {
               case ConnectionEstablished :

                  OnConnectionEstablished(socket.GetRef());

               break;

               case ReadCompleted :
               {
                  CSmartBuffer buffer(static_cast<IBuffer *>(pWorkItem));

                  ProcessMessage(socket.GetRef(), buffer);
               }
               break;

               case Timer :

                  OnTimer(socket.GetRef());

               break;
            }
         }
         catch(...)
         {
            socket->AbortConnection();
            throw;
         }
      }
   }
   catch(const CException &e)
   {
      OutputEx(_T("Process - Exception - ") + e.GetDetails());
   }
   catch(...)
   {
      OutputEx(_T("Process - Unexpected exception"));
   }
}

void CThreadPool::Shutdown()
{
   Output(_T("Shutdown"));
}

void CThreadPool::OnThreadPoolThreadStarted()
{
   Output(_T("OnThreadStarted"));
}

void CThreadPool::OnThreadPoolDispatch()
{
}

void CThreadPool::OnThreadPoolProcessingStarted(
   const bool /*dataAvailable*/)
{
   Output(_T("OnProcessingStarted"));
}

void CThreadPool::OnThreadPoolProcessingStopped(
   const bool /*dataAvailable*/)
{
   Output(_T("OnProcessingStopped"));
}

void CThreadPool::OnThreadPoolThreadStopped()
{
   Output(_T("OnThreadStopped"));
}

///////////////////////////////////////////////////////////////////////////////
// CThreadPool::CPerConnectionData
///////////////////////////////////////////////////////////////////////////////

CThreadPool::CPerConnectionData::CPerConnectionData(
   const _tstring &address,
   IStreamSocket &socket,
   IQueueTimers &timerQueue,
   CThreadPool &pool)
   :  m_numUses(0),
      m_address(address),
      m_socket(socket),
      m_timerQueue(timerQueue),
      m_timerHandle(timerQueue.CreateTimer()),
      m_pool(pool)
{
   SetPeriodicTimer();
}

CThreadPool::CPerConnectionData::~CPerConnectionData()
{
   m_timerQueue.DestroyTimer(m_timerHandle);
}

void CThreadPool::CPerConnectionData::SetPeriodicTimer()
{
   static const Milliseconds connectionActivityDelay = 5000;   // 5 seconds...

   m_socket.AddRef();      // Take a reference to the socket whilst we are held by the timer queue
                           // this will keep us alive until the timer is triggered or cancelled

   m_timerQueue.SetTimer(m_timerHandle, *this, connectionActivityDelay, reinterpret_cast<UserData>(this));
}

void CThreadPool::CPerConnectionData::CancelTimer()
{
   if (m_timerQueue.CancelTimer(m_timerHandle))
   {
      m_socket.Release();  // The timer was active when it was cancelled, release the reference
                           // that we took in SetPeriodicTimer()
   }
}

void CThreadPool::CPerConnectionData::UseConnection()
{
   ::InterlockedIncrement(&m_numUses);
}

_tstring CThreadPool::CPerConnectionData::GetConnectionDetails() const
{
   return m_address + _T(" - ") + ToString(m_numUses);
}

void CThreadPool::CPerConnectionData::OnTimer(
   UserData userData)
{
   CPerConnectionData &data = *reinterpret_cast<CPerConnectionData *>(userData);

   data.OnTimer();
}

void CThreadPool::CPerConnectionData::OnTimer()
{
   m_pool.DispatchTimer(m_socket);

   m_socket.Release();     // The timer queue can no longer access the user data
                           // so we can release the reference on the socket that
                           // we took in SetPeriodicTimer()
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ThreadPool.cpp
///////////////////////////////////////////////////////////////////////////////
