#pragma once
///////////////////////////////////////////////////////////////////////////////
// File: ThreadPool.h
/////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2009 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/CoreTools/ThreadPool.h"
#include "JetByteTools/CoreTools/NullThreadPoolMonitor.h"
#include "JetByteTools/CoreTools/IIndexedOpaqueUserData.h"
#include "JetByteTools/CoreTools/IQueueTimers.h"
#include "JetByteTools/CoreTools/IThreadPoolWorkerThreadFactory.h"
#include "JetByteTools/CoreTools/IThreadPoolWorkerThreadCallback.h"

#include "JetByteTools/IOTools/SmartBuffer.h"

///////////////////////////////////////////////////////////////////////////////
// Classes defined in other files...
///////////////////////////////////////////////////////////////////////////////

namespace JetByteTools
{
   namespace Core
   {
      class CEvent;
      class IProvideUserData;
   }

   namespace IO
   {
      class IBuffer;
      class IAllocateBuffers;
   }

   namespace Socket
   {
      class IStreamSocket;
      class IAddress;
      class CStreamSocketBroadcastableConnectionCollection;
   }
}

///////////////////////////////////////////////////////////////////////////////
// CThreadPool
///////////////////////////////////////////////////////////////////////////////

class CThreadPool :
   private JetByteTools::Core::IThreadPoolWorkerThreadFactory,
   private JetByteTools::Core::IThreadPoolWorkerThreadCallback,
   private JetByteTools::Core::CNullThreadPoolMonitor
{
   public:

      typedef JetByteTools::Core::CThreadPool::ThreadCount ThreadCount;

      CThreadPool(
         JetByteTools::Core::IQueueTimers &timerQueue,
         JetByteTools::Socket::CStreamSocketBroadcastableConnectionCollection &connections,
         JetByteTools::IO::IAllocateBuffers &bufferAllocator,
         JetByteTools::Core::IProvideUserData &userDataProvider,
         ThreadCount initialThreads,
         ThreadCount minThreads,
         ThreadCount maxThreads,
         ThreadCount maxDormantThreads,
         JetByteTools::Milliseconds poolMaintPeriod,
         JetByteTools::Milliseconds dispatchTimeout);

      CThreadPool(
         const CThreadPool &rhs) = delete;

      CThreadPool &operator=(
         const CThreadPool &rhs) = delete;

      void Start();

      void WaitForShutdownToComplete(
         JetByteTools::Milliseconds forceTerminationAfter = INFINITE);

      void DispatchConnectionEstablished(
         JetByteTools::Socket::IStreamSocket &socket,
         const JetByteTools::Socket::IAddress &address);

      void DispatchReadCompleted(
         JetByteTools::Socket::IStreamSocket &socket,
         JetByteTools::IO::CSmartBuffer &buffer);

      void DispatchTimer(
         JetByteTools::Socket::IStreamSocket &socket);

      void DispatchServerTimer(
         DWORD timerId);

      void OnSocketReleased(
         JetByteTools::Core::IIndexedOpaqueUserData &userData);

      void OnConnectionClientClose(
          JetByteTools::Socket::IStreamSocket &socket);

      void OnConnectionReset(
          JetByteTools::Socket::IStreamSocket &socket,
          DWORD lastError);

   private :

      enum DispatchEvents
      {
         ConnectionEstablished,
         ReadCompleted,
         Timer,
         ServerTimer
      };

      class CPerConnectionData : private JetByteTools::Core::IQueueTimers::Timer
      {
         public :

            CPerConnectionData(
               const JetByteTools::Core::_tstring &address,
               JetByteTools::Socket::IStreamSocket &socket,
               JetByteTools::Core::IQueueTimers &timerQueue,
               CThreadPool &pool);

            CPerConnectionData(
               const CPerConnectionData &rhs) = delete;

            ~CPerConnectionData();

            CPerConnectionData &operator=(
               const CPerConnectionData &rhs) = delete;

            void UseConnection();

            JetByteTools::Core::_tstring GetConnectionDetails() const;

            void SetPeriodicTimer();

            void CancelTimer();

         private :

            void OnTimer();

            // Implement IQueueTimers::Timer

            void OnTimer(
               UserData userData) override;

            long m_numUses;

            const JetByteTools::Core::_tstring m_address;

            JetByteTools::Socket::IStreamSocket &m_socket;

            JetByteTools::Core::IQueueTimers &m_timerQueue;

            JetByteTools::Core::IQueueTimers::Handle m_timerHandle;

            CThreadPool &m_pool;
      };

      CPerConnectionData &GetPerConnectionData(
         const JetByteTools::Core::IIndexedOpaqueUserData &userData) const;

      void DoDispatch(
         JetByteTools::Socket::IStreamSocket &socket,
         JetByteTools::IO::IBuffer *pBuffer,
         DispatchEvents event);

      void ProcessMessage(
         JetByteTools::Socket::IStreamSocket &socket,
         const JetByteTools::IO::CSmartBuffer &buffer) const;

      void OnConnectionEstablished(
         JetByteTools::Socket::IStreamSocket &socket);

      void OnTimer(
         JetByteTools::Socket::IStreamSocket &socket);

      void OnServerTimer(
         DWORD timerId);

      // Implement IThreadPoolWorkerThreadFactory

      JetByteTools::Core::IThreadPoolWorkerThreadCallback &CreateWorkerThread() override;

      void DestroyWorkerThread(
         JetByteTools::Core::IThreadPoolWorkerThreadCallback &thread) override;

      void DisposeOfQueuedItemsAfterShutdown(
         ULONG_PTR completionKey,
         DWORD numberOfBytes,
         JetByteTools::Core::IAsyncWorkItem *pWorkItem) override;

      // Implement IIOCPWorkerThreadCallback

      bool Initialise() override;

      void Process(
         ULONG_PTR completionKey,
         DWORD numberOfBytes,
         JetByteTools::Core::IAsyncWorkItem *pWorkItem) override;

      void Shutdown() override;

      void OnProcessException(
         const JetByteTools::Core::_tstring & /*message*/) override {}

      void OnThreadTerminationException(
         const JetByteTools::Core::_tstring & /*message*/) override {}

      // Implement IMonitorThreadPool

      void OnThreadPoolThreadStarted() override;

      void OnThreadPoolDispatch() override;

      void OnThreadPoolProcessingStarted(
         bool dataAvailable) override;

      void OnThreadPoolProcessingStopped(
         bool dataAvailable) override;

      void OnThreadPoolThreadStopped() override;

      JetByteTools::Core::CThreadPool m_threadPool;

      JetByteTools::Core::IQueueTimers &m_timerQueue;

      JetByteTools::Socket::CStreamSocketBroadcastableConnectionCollection &m_connections;

      JetByteTools::IO::IAllocateBuffers &m_bufferAllocator;

      const JetByteTools::Core::IIndexedOpaqueUserData::UserDataIndex m_userDataIndex;
};

///////////////////////////////////////////////////////////////////////////////
// End of file: ThreadPool.h
///////////////////////////////////////////////////////////////////////////////
